#include "../../version.h"
#if (defined(RL_TOOLS_DISABLE_INCLUDE_GUARDS) || !defined(RL_TOOLS_OPERATIONS_CPU_GROUP_2_H)) && (RL_TOOLS_USE_THIS_VERSION == 1)
#pragma once
#define RL_TOOLS_OPERATIONS_CPU_GROUP_2_H
#if defined(RL_TOOLS_DISABLE_INCLUDE_GUARDS) || !defined(RL_TOOLS_OPERATIONS_CPU_GROUP_2)
    #define RL_TOOLS_OPERATIONS_CPU_GROUP_2
    #include "../../utils/assert/operations_cpu.h"
    #include "../../random/operations_cpu.h"
#else
    #error "Group 2 already imported"
#endif
#endif
